@@ -3,6 +3,13 @@
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
 
+"""Conversion from filesystem tree to SWH objects.
+
+This module allows reading a tree of directories and files from a local
+filesystem, and convert them to in-memory data structures, which can then
+be exported to SWH data model objects, as defined in :mod:`swh.model.model`.
+"""
+
 import datetime
 import enum
 import fnmatch
@@ -18,11 +25,10 @@ from typing_extensions import Final
 
 from . import model
 from .exceptions import InvalidDirectoryPath
-from .hashutil import MultiHash
-from .identifiers import directory_entry_sort_key, directory_identifier
-from .identifiers import identifier_to_bytes as id_to_bytes
-from .identifiers import identifier_to_str as id_to_str
+from .git_objects import directory_entry_sort_key
+from .hashutil import MultiHash, hash_to_hex
 from .merkle import MerkleLeaf, MerkleNode
+from .swhids import CoreSWHID, ObjectType
 
 
 @attr.s(frozen=True, slots=True)
@@ -207,8 +213,13 @@ class Content(MerkleLeaf):
         obj = cls(ret)
         return obj
 
+    def swhid(self) -> CoreSWHID:
+        """Return node identifier as a SWHID
+        """
+        return CoreSWHID(object_type=ObjectType.CONTENT, object_id=self.hash)
+
     def __repr__(self):
-        return "Content(id=%s)" % id_to_str(self.hash)
+        return "Content(id=%s)" % hash_to_hex(self.hash)
 
     def compute_hash(self):
         return self.data["sha1_git"]
@@ -429,7 +440,7 @@ class Directory(MerkleNode):
                     if dir_filter(path, name, dirs[path].entries):
                         entries[name] = dirs[path]
 
-            dirs[root] = cls({"name": os.path.basename(root)})
+            dirs[root] = cls({"name": os.path.basename(root), "path": root})
             dirs[root].update(entries)
 
         return dirs[top_path]
@@ -469,8 +480,8 @@ class Directory(MerkleNode):
 
     @property
     def entries(self):
-        """Child nodes, sorted by name in the same way `directory_identifier`
-        does."""
+        """Child nodes, sorted by name in the same way
+        :func:`swh.model.git_objects.directory_git_object` does."""
         if self.__entries is None:
             self.__entries = sorted(
                 (
@@ -482,8 +493,13 @@ class Directory(MerkleNode):
 
         return self.__entries
 
+    def swhid(self) -> CoreSWHID:
+        """Return node identifier as a SWHID
+        """
+        return CoreSWHID(object_type=ObjectType.DIRECTORY, object_id=self.hash)
+
     def compute_hash(self):
-        return id_to_bytes(directory_identifier({"entries": self.entries}))
+        return model.Directory.from_dict({"entries": self.entries}).id
 
     def to_model(self) -> model.Directory:
         """Builds a `model.Directory` object based on this node;
@@ -533,8 +549,15 @@ class Directory(MerkleNode):
             key1, key2 = key.rsplit(b"/", 1)
             del self[key1][key2]
 
+    def __contains__(self, key):
+        if b"/" not in key:
+            return super().__contains__(key)
+        else:
+            key1, key2 = key.split(b"/", 1)
+            return super().__contains__(key1) and self[key1].__contains__(key2)
+
     def __repr__(self):
         return "Directory(id=%s, entries=[%s])" % (
-            id_to_str(self.hash),
+            hash_to_hex(self.hash),
             ", ".join(str(entry) for entry in self),
         )
