@@ -3,34 +3,37 @@
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
 
+"""
+Implementation of Software Heritage's data model
+
+See :ref:`data-model` for an overview of the data model.
+
+The classes defined in this module are immutable
+`attrs objects <https://attrs.org/>`__ and enums.
+
+All classes define a ``from_dict`` class method and a ``to_dict``
+method to convert between them and msgpack-serializable objects.
+"""
+
 from abc import ABCMeta, abstractmethod
 import datetime
 from enum import Enum
-from hashlib import sha256
+import hashlib
 from typing import Any, Dict, Iterable, Optional, Tuple, TypeVar, Union
 
 import attr
-from attrs_strict import type_validator
+from attrs_strict import AttributeTypeError
 import dateutil.parser
 import iso8601
 from typing_extensions import Final
 
+from . import git_objects
 from .collections import ImmutableDict
-from .hashutil import DEFAULT_ALGORITHMS, MultiHash, hash_to_bytes
-from .identifiers import (
-    directory_identifier,
-    extid_identifier,
-    normalize_timestamp,
-    origin_identifier,
-    raw_extrinsic_metadata_identifier,
-    release_identifier,
-    revision_identifier,
-    snapshot_identifier,
-)
-from .identifiers import CoreSWHID
-from .identifiers import ExtendedObjectType as SwhidExtendedObjectType
-from .identifiers import ExtendedSWHID
-from .identifiers import ObjectType as SwhidObjectType
+from .hashutil import DEFAULT_ALGORITHMS, MultiHash, hash_to_hex
+from .swhids import CoreSWHID
+from .swhids import ExtendedObjectType as SwhidExtendedObjectType
+from .swhids import ExtendedSWHID
+from .swhids import ObjectType as SwhidObjectType
 
 
 class MissingData(Exception):
@@ -46,6 +49,8 @@ KeyType = Union[Dict[str, str], Dict[str, bytes], bytes]
 
 SHA1_SIZE = 20
 
+_OFFSET_CHARS = frozenset(b"+-0123456789")
+
 # TODO: Limit this to 20 bytes
 Sha1Git = bytes
 Sha1 = bytes
@@ -55,6 +60,13 @@ KT = TypeVar("KT")
 VT = TypeVar("VT")
 
 
+def hash_repr(h: bytes) -> str:
+    if h is None:
+        return "None"
+    else:
+        return f"hash_to_bytes('{hash_to_hex(h)}')"
+
+
 def freeze_optional_dict(
     d: Union[None, Dict[KT, VT], ImmutableDict[KT, VT]]  # type: ignore
 ) -> Optional[ImmutableDict[KT, VT]]:
@@ -80,6 +92,74 @@ def dictify(value):
         return value
 
 
+def _check_type(type_, value):
+    if type_ is object or type_ is Any:
+        return True
+
+    if type_ is None:
+        return value is None
+
+    origin = getattr(type_, "__origin__", None)
+
+    # Non-generic type, check it directly
+    if origin is None:
+        # This is functionally equivalent to using just this:
+        #   return isinstance(value, type)
+        # but using type equality before isinstance allows very quick checks
+        # when the exact class is used (which is the overwhelming majority of cases)
+        # while still allowing subclasses to be used.
+        return type(value) == type_ or isinstance(value, type_)
+
+    # Check the type of the value itself
+    #
+    # For the same reason as above, this condition is functionally equivalent to:
+    #   if origin is not Union and not isinstance(value, origin):
+    if origin is not Union and type(value) != origin and not isinstance(value, origin):
+        return False
+
+    # Then, if it's a container, check its items.
+    if origin is tuple:
+        args = type_.__args__
+        if len(args) == 2 and args[1] is Ellipsis:
+            # Infinite tuple
+            return all(_check_type(args[0], item) for item in value)
+        else:
+            # Finite tuple
+            if len(args) != len(value):
+                return False
+
+            return all(
+                _check_type(item_type, item) for (item_type, item) in zip(args, value)
+            )
+    elif origin is Union:
+        args = type_.__args__
+        return any(_check_type(variant, value) for variant in args)
+    elif origin is ImmutableDict:
+        (key_type, value_type) = type_.__args__
+        return all(
+            _check_type(key_type, key) and _check_type(value_type, value)
+            for (key, value) in value.items()
+        )
+    else:
+        # No need to check dict or list. because they are converted to ImmutableDict
+        # and tuple respectively.
+        raise NotImplementedError(f"Type-checking {type_}")
+
+
+def type_validator():
+    """Like attrs_strict.type_validator(), but stricter.
+
+    It is an attrs validator, which checks attributes have the specified type,
+    using type equality instead of ``isinstance()``, for improved performance
+    """
+
+    def validator(instance, attribute, value):
+        if not _check_type(attribute.type, value):
+            raise AttributeTypeError(value, attribute)
+
+    return validator
+
+
 ModelType = TypeVar("ModelType", bound="BaseModel")
 
 
@@ -114,6 +194,14 @@ class BaseModel:
         deduplication."""
         raise NotImplementedError(f"unique_key for {self}")
 
+    def check(self) -> None:
+        """Performs internal consistency checks, and raises an error if one fails."""
+        attr.validate(self)
+
+
+def _compute_hash_from_manifest(manifest: bytes) -> Sha1Git:
+    return hashlib.new("sha1", manifest).digest()
+
 
 class HashableObject(metaclass=ABCMeta):
     """Mixin to automatically compute object identifier hash when
@@ -121,7 +209,8 @@ class HashableObject(metaclass=ABCMeta):
 
     __slots__ = ()
 
-    @abstractmethod
+    id: Sha1Git
+
     def compute_hash(self) -> bytes:
         """Derived model classes must implement this to compute
         the object hash.
@@ -129,7 +218,11 @@ class HashableObject(metaclass=ABCMeta):
         This method is called by the object initialization if the `id`
         attribute is set to an empty value.
         """
-        pass
+        return self._compute_hash_from_attributes()
+
+    @abstractmethod
+    def _compute_hash_from_attributes(self) -> Sha1Git:
+        raise NotImplementedError(f"_compute_hash_from_attributes for {self}")
 
     def __attrs_post_init__(self):
         if not self.id:
@@ -137,7 +230,54 @@ class HashableObject(metaclass=ABCMeta):
             object.__setattr__(self, "id", obj_id)
 
     def unique_key(self) -> KeyType:
-        return self.id  # type: ignore
+        return self.id
+
+    def check(self) -> None:
+        super().check()  # type: ignore
+
+        if self.id != self.compute_hash():
+            raise ValueError("'id' does not match recomputed hash.")
+
+
+class HashableObjectWithManifest(HashableObject):
+    """Derived class of HashableObject, for objects that may need to store
+    verbatim git objects as ``raw_manifest`` to preserve original hashes."""
+
+    raw_manifest: Optional[bytes] = None
+    """Stores the original content of git objects when they cannot be faithfully
+    represented using only the other attributes.
+
+    This should only be used as a last resort, and only set in the Git loader,
+    for objects too corrupt to fit the data model."""
+
+    def to_dict(self):
+        d = super().to_dict()
+        if d["raw_manifest"] is None:
+            del d["raw_manifest"]
+        return d
+
+    def compute_hash(self) -> bytes:
+        """Derived model classes must implement this to compute
+        the object hash.
+
+        This method is called by the object initialization if the `id`
+        attribute is set to an empty value.
+        """
+        if self.raw_manifest is None:
+            return super().compute_hash()
+        else:
+            return _compute_hash_from_manifest(self.raw_manifest)
+
+    def check(self) -> None:
+        super().check()
+
+        if (
+            self.raw_manifest is not None
+            and self.id == self._compute_hash_from_attributes()
+        ):
+            raise ValueError(
+                f"{self} has a non-none raw_manifest attribute, but does not need it."
+            )
 
 
 @attr.s(frozen=True, slots=True)
@@ -191,7 +331,29 @@ class Person(BaseModel):
         Anonymization is simply a Person which fullname is the hashed, with unset name
         or email.
         """
-        return Person(fullname=sha256(self.fullname).digest(), name=None, email=None,)
+        return Person(
+            fullname=hashlib.sha256(self.fullname).digest(), name=None, email=None,
+        )
+
+    @classmethod
+    def from_dict(cls, d):
+        """
+        If the fullname is missing, construct a fullname
+        using the following heuristics: if the name value is None, we return the
+        email in angle brackets, else, we return the name, a space, and the email
+        in angle brackets.
+        """
+        if "fullname" not in d:
+            parts = []
+            if d["name"] is not None:
+                parts.append(d["name"])
+            if d["email"] is not None:
+                parts.append(b"".join([b"<", d["email"], b">"]))
+
+            fullname = b" ".join(parts)
+            d = {**d, "fullname": fullname}
+        d = {"name": None, "email": None, **d}
+        return super().from_dict(d)
 
 
 @attr.s(frozen=True, slots=True)
@@ -226,6 +388,15 @@ class TimestampWithTimezone(BaseModel):
     offset = attr.ib(type=int, validator=type_validator())
     negative_utc = attr.ib(type=bool, validator=type_validator())
 
+    offset_bytes = attr.ib(type=bytes, validator=type_validator())
+    """Raw git representation of the timezone, as an offset from UTC.
+    It should follow this format: ``+HHMM`` or ``-HHMM`` (including ``+0000`` and
+    ``-0000``).
+
+    However, when created from git objects, it must be the exact bytes used in the
+    original objects, so it may differ from this format when they do.
+    """
+
     @offset.validator
     def check_offset(self, attribute, value):
         """Checks the offset is a 16-bits signed integer (in theory, it
@@ -235,28 +406,123 @@ class TimestampWithTimezone(BaseModel):
             # you'll find in the wild...
             raise ValueError("offset too large: %d minutes" % value)
 
+        self._check_offsets_match()
+
     @negative_utc.validator
     def check_negative_utc(self, attribute, value):
         if self.offset and value:
             raise ValueError("negative_utc can only be True is offset=0")
 
+        self._check_offsets_match()
+
+    @offset_bytes.default
+    def _default_offset_bytes(self):
+        negative = self.offset < 0 or self.negative_utc
+        (hours, minutes) = divmod(abs(self.offset), 60)
+        return f"{'-' if negative else '+'}{hours:02}{minutes:02}".encode()
+
+    @offset_bytes.validator
+    def check_offset_bytes(self, attribute, value):
+        if not set(value) <= _OFFSET_CHARS:
+            raise ValueError(f"invalid characters in offset_bytes: {value!r}")
+
+        self._check_offsets_match()
+
+    def _check_offsets_match(self):
+        offset_str = self.offset_bytes.decode()
+        assert offset_str[0] in "+-"
+        sign = int(offset_str[0] + "1")
+        hours = int(offset_str[1:-2])
+        minutes = int(offset_str[-2:])
+        offset = sign * (hours * 60 + minutes)
+        if offset != self.offset:
+            raise ValueError(
+                f"offset_bytes ({self.offset_bytes!r}) does not match offset "
+                f"{divmod(self.offset, 60)}"
+            )
+
+        if offset == 0 and self.negative_utc != self.offset_bytes.startswith(b"-"):
+            raise ValueError(
+                f"offset_bytes ({self.offset_bytes!r}) does not match negative_utc "
+                f"({self.negative_utc})"
+            )
+
     @classmethod
-    def from_dict(cls, obj: Union[Dict, datetime.datetime, int]):
+    def from_dict(cls, time_representation: Union[Dict, datetime.datetime, int]):
         """Builds a TimestampWithTimezone from any of the formats
         accepted by :func:`swh.model.normalize_timestamp`."""
         # TODO: this accept way more types than just dicts; find a better
         # name
-        d = normalize_timestamp(obj)
+        negative_utc = False
+
+        if isinstance(time_representation, dict):
+            ts = time_representation["timestamp"]
+            if isinstance(ts, dict):
+                seconds = ts.get("seconds", 0)
+                microseconds = ts.get("microseconds", 0)
+            elif isinstance(ts, int):
+                seconds = ts
+                microseconds = 0
+            else:
+                raise ValueError(
+                    f"TimestampWithTimezone.from_dict received non-integer timestamp "
+                    f"member {ts!r}"
+                )
+            offset = time_representation["offset"]
+            if "negative_utc" in time_representation:
+                negative_utc = time_representation["negative_utc"]
+            if negative_utc is None:
+                negative_utc = False
+        elif isinstance(time_representation, datetime.datetime):
+            microseconds = time_representation.microsecond
+            if microseconds:
+                time_representation = time_representation.replace(microsecond=0)
+            seconds = int(time_representation.timestamp())
+            utcoffset = time_representation.utcoffset()
+            if utcoffset is None:
+                raise ValueError(
+                    f"TimestampWithTimezone.from_dict received datetime without "
+                    f"timezone: {time_representation}"
+                )
+
+            # utcoffset is an integer number of minutes
+            seconds_offset = utcoffset.total_seconds()
+            offset = int(seconds_offset) // 60
+        elif isinstance(time_representation, int):
+            seconds = time_representation
+            microseconds = 0
+            offset = 0
+        else:
+            raise ValueError(
+                f"TimestampWithTimezone.from_dict received non-integer timestamp: "
+                f"{time_representation!r}"
+            )
+
         return cls(
-            timestamp=Timestamp.from_dict(d["timestamp"]),
-            offset=d["offset"],
-            negative_utc=d["negative_utc"],
+            timestamp=Timestamp(seconds=seconds, microseconds=microseconds),
+            offset=offset,
+            negative_utc=negative_utc,
         )
 
     @classmethod
     def from_datetime(cls, dt: datetime.datetime):
         return cls.from_dict(dt)
 
+    def to_datetime(self) -> datetime.datetime:
+        """Convert to a datetime (with a timezone set to the recorded fixed UTC offset)
+
+        Beware that this conversion can be lossy: the negative_utc flag is not
+        taken into consideration (since it cannot be represented in a
+        datetime). Also note that it may fail due to type overflow.
+
+        """
+        timestamp = datetime.datetime.fromtimestamp(
+            self.timestamp.seconds,
+            datetime.timezone(datetime.timedelta(minutes=self.offset)),
+        )
+        timestamp = timestamp.replace(microsecond=self.timestamp.microseconds)
+        return timestamp
+
     @classmethod
     def from_iso8601(cls, s):
         """Builds a TimestampWithTimezone from an ISO8601-formatted string.
@@ -264,26 +530,31 @@ class TimestampWithTimezone(BaseModel):
         dt = iso8601.parse_date(s)
         tstz = cls.from_datetime(dt)
         if dt.tzname() == "-00:00":
-            tstz = attr.evolve(tstz, negative_utc=True)
+            assert tstz.offset_bytes == b"+0000"
+            tstz = attr.evolve(tstz, negative_utc=True, offset_bytes=b"-0000")
         return tstz
 
 
 @attr.s(frozen=True, slots=True)
-class Origin(BaseModel):
+class Origin(HashableObject, BaseModel):
     """Represents a software source: a VCS and an URL."""
 
     object_type: Final = "origin"
 
     url = attr.ib(type=str, validator=type_validator())
 
+    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"")
+
     def unique_key(self) -> KeyType:
         return {"url": self.url}
 
+    def _compute_hash_from_attributes(self) -> bytes:
+        return _compute_hash_from_manifest(self.url.encode("utf-8"))
+
     def swhid(self) -> ExtendedSWHID:
         """Returns a SWHID representing this origin."""
         return ExtendedSWHID(
-            object_type=SwhidExtendedObjectType.ORIGIN,
-            object_id=hash_to_bytes(origin_identifier(self.unique_key())),
+            object_type=SwhidExtendedObjectType.ORIGIN, object_id=self.id,
         )
 
 
@@ -336,7 +607,9 @@ class OriginVisitStatus(BaseModel):
             ["created", "ongoing", "full", "partial", "not_found", "failed"]
         ),
     )
-    snapshot = attr.ib(type=Optional[Sha1Git], validator=type_validator())
+    snapshot = attr.ib(
+        type=Optional[Sha1Git], validator=type_validator(), repr=hash_repr
+    )
     # Type is optional be to able to use it before adding it to the database model
     type = attr.ib(type=Optional[str], validator=type_validator(), default=None)
     metadata = attr.ib(
@@ -367,6 +640,9 @@ class TargetType(Enum):
     SNAPSHOT = "snapshot"
     ALIAS = "alias"
 
+    def __repr__(self):
+        return f"TargetType.{self.name}"
+
 
 class ObjectType(Enum):
     """The type of content pointed to by a release. Usually a revision"""
@@ -377,6 +653,9 @@ class ObjectType(Enum):
     RELEASE = "release"
     SNAPSHOT = "snapshot"
 
+    def __repr__(self):
+        return f"ObjectType.{self.name}"
+
 
 @attr.s(frozen=True, slots=True)
 class SnapshotBranch(BaseModel):
@@ -384,7 +663,7 @@ class SnapshotBranch(BaseModel):
 
     object_type: Final = "snapshot_branch"
 
-    target = attr.ib(type=bytes, validator=type_validator())
+    target = attr.ib(type=bytes, validator=type_validator(), repr=hash_repr)
     target_type = attr.ib(type=TargetType, validator=type_validator())
 
     @target.validator
@@ -411,10 +690,10 @@ class Snapshot(HashableObject, BaseModel):
         validator=type_validator(),
         converter=freeze_optional_dict,
     )
-    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"")
+    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"", repr=hash_repr)
 
-    def compute_hash(self) -> bytes:
-        return hash_to_bytes(snapshot_identifier(self.to_dict()))
+    def _compute_hash_from_attributes(self) -> bytes:
+        return _compute_hash_from_manifest(git_objects.snapshot_git_object(self))
 
     @classmethod
     def from_dict(cls, d):
@@ -433,12 +712,12 @@ class Snapshot(HashableObject, BaseModel):
 
 
 @attr.s(frozen=True, slots=True)
-class Release(HashableObject, BaseModel):
+class Release(HashableObjectWithManifest, BaseModel):
     object_type: Final = "release"
 
     name = attr.ib(type=bytes, validator=type_validator())
     message = attr.ib(type=Optional[bytes], validator=type_validator())
-    target = attr.ib(type=Optional[Sha1Git], validator=type_validator())
+    target = attr.ib(type=Optional[Sha1Git], validator=type_validator(), repr=hash_repr)
     target_type = attr.ib(type=ObjectType, validator=type_validator())
     synthetic = attr.ib(type=bool, validator=type_validator())
     author = attr.ib(type=Optional[Person], validator=type_validator(), default=None)
@@ -451,10 +730,11 @@ class Release(HashableObject, BaseModel):
         converter=freeze_optional_dict,
         default=None,
     )
-    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"")
+    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"", repr=hash_repr)
+    raw_manifest = attr.ib(type=Optional[bytes], default=None)
 
-    def compute_hash(self) -> bytes:
-        return hash_to_bytes(release_identifier(self.to_dict()))
+    def _compute_hash_from_attributes(self) -> bytes:
+        return _compute_hash_from_manifest(git_objects.release_git_object(self))
 
     @author.validator
     def check_author(self, attribute, value):
@@ -496,6 +776,11 @@ class RevisionType(Enum):
     DSC = "dsc"
     SUBVERSION = "svn"
     MERCURIAL = "hg"
+    CVS = "cvs"
+    BAZAAR = "bzr"
+
+    def __repr__(self):
+        return f"RevisionType.{self.name}"
 
 
 def tuplify_extra_headers(value: Iterable):
@@ -503,7 +788,7 @@ def tuplify_extra_headers(value: Iterable):
 
 
 @attr.s(frozen=True, slots=True)
-class Revision(HashableObject, BaseModel):
+class Revision(HashableObjectWithManifest, BaseModel):
     object_type: Final = "revision"
 
     message = attr.ib(type=Optional[bytes], validator=type_validator())
@@ -514,7 +799,7 @@ class Revision(HashableObject, BaseModel):
         type=Optional[TimestampWithTimezone], validator=type_validator()
     )
     type = attr.ib(type=RevisionType, validator=type_validator())
-    directory = attr.ib(type=Sha1Git, validator=type_validator())
+    directory = attr.ib(type=Sha1Git, validator=type_validator(), repr=hash_repr)
     synthetic = attr.ib(type=bool, validator=type_validator())
     metadata = attr.ib(
         type=Optional[ImmutableDict[str, object]],
@@ -523,13 +808,14 @@ class Revision(HashableObject, BaseModel):
         default=None,
     )
     parents = attr.ib(type=Tuple[Sha1Git, ...], validator=type_validator(), default=())
-    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"")
+    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"", repr=hash_repr)
     extra_headers = attr.ib(
         type=Tuple[Tuple[bytes, bytes], ...],
         validator=type_validator(),
         converter=tuplify_extra_headers,
         default=(),
     )
+    raw_manifest = attr.ib(type=Optional[bytes], default=None)
 
     def __attrs_post_init__(self):
         super().__attrs_post_init__()
@@ -545,8 +831,8 @@ class Revision(HashableObject, BaseModel):
                 attr.validate(self)
             object.__setattr__(self, "metadata", metadata)
 
-    def compute_hash(self) -> bytes:
-        return hash_to_bytes(revision_identifier(self.to_dict()))
+    def _compute_hash_from_attributes(self) -> bytes:
+        return _compute_hash_from_manifest(git_objects.revision_git_object(self))
 
     @classmethod
     def from_dict(cls, d):
@@ -590,20 +876,36 @@ class DirectoryEntry(BaseModel):
 
     name = attr.ib(type=bytes, validator=type_validator())
     type = attr.ib(type=str, validator=attr.validators.in_(["file", "dir", "rev"]))
-    target = attr.ib(type=Sha1Git, validator=type_validator())
-    perms = attr.ib(type=int, validator=type_validator())
+    target = attr.ib(type=Sha1Git, validator=type_validator(), repr=hash_repr)
+    perms = attr.ib(type=int, validator=type_validator(), converter=int, repr=oct)
     """Usually one of the values of `swh.model.from_disk.DentryPerms`."""
 
+    @name.validator
+    def check_name(self, attribute, value):
+        if b"/" in value:
+            raise ValueError(f"{value!r} is not a valid directory entry name.")
+
 
 @attr.s(frozen=True, slots=True)
-class Directory(HashableObject, BaseModel):
+class Directory(HashableObjectWithManifest, BaseModel):
     object_type: Final = "directory"
 
     entries = attr.ib(type=Tuple[DirectoryEntry, ...], validator=type_validator())
-    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"")
+    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"", repr=hash_repr)
+    raw_manifest = attr.ib(type=Optional[bytes], default=None)
 
-    def compute_hash(self) -> bytes:
-        return hash_to_bytes(directory_identifier(self.to_dict()))
+    def _compute_hash_from_attributes(self) -> bytes:
+        return _compute_hash_from_manifest(git_objects.directory_git_object(self))
+
+    @entries.validator
+    def check_entries(self, attribute, value):
+        seen = set()
+        for entry in value:
+            if entry.name in seen:
+                raise ValueError(
+                    "{self.swhid()} has duplicated entry name: {entry.name!r}"
+                )
+            seen.add(entry.name)
 
     @classmethod
     def from_dict(cls, d):
@@ -660,10 +962,10 @@ class BaseContent(BaseModel):
 class Content(BaseContent):
     object_type: Final = "content"
 
-    sha1 = attr.ib(type=bytes, validator=type_validator())
-    sha1_git = attr.ib(type=Sha1Git, validator=type_validator())
-    sha256 = attr.ib(type=bytes, validator=type_validator())
-    blake2s256 = attr.ib(type=bytes, validator=type_validator())
+    sha1 = attr.ib(type=bytes, validator=type_validator(), repr=hash_repr)
+    sha1_git = attr.ib(type=Sha1Git, validator=type_validator(), repr=hash_repr)
+    sha256 = attr.ib(type=bytes, validator=type_validator(), repr=hash_repr)
+    blake2s256 = attr.ib(type=bytes, validator=type_validator(), repr=hash_repr)
 
     length = attr.ib(type=int, validator=type_validator())
 
@@ -743,10 +1045,14 @@ class Content(BaseContent):
 class SkippedContent(BaseContent):
     object_type: Final = "skipped_content"
 
-    sha1 = attr.ib(type=Optional[bytes], validator=type_validator())
-    sha1_git = attr.ib(type=Optional[Sha1Git], validator=type_validator())
-    sha256 = attr.ib(type=Optional[bytes], validator=type_validator())
-    blake2s256 = attr.ib(type=Optional[bytes], validator=type_validator())
+    sha1 = attr.ib(type=Optional[bytes], validator=type_validator(), repr=hash_repr)
+    sha1_git = attr.ib(
+        type=Optional[Sha1Git], validator=type_validator(), repr=hash_repr
+    )
+    sha256 = attr.ib(type=Optional[bytes], validator=type_validator(), repr=hash_repr)
+    blake2s256 = attr.ib(
+        type=Optional[bytes], validator=type_validator(), repr=hash_repr
+    )
 
     length = attr.ib(type=Optional[int], validator=type_validator())
 
@@ -824,6 +1130,9 @@ class MetadataAuthorityType(Enum):
     FORGE = "forge"
     REGISTRY = "registry"
 
+    def __repr__(self):
+        return f"MetadataAuthorityType.{self.name}"
+
 
 @attr.s(frozen=True, slots=True)
 class MetadataAuthority(BaseModel):
@@ -885,6 +1194,17 @@ class MetadataFetcher(BaseModel):
         return {"name": self.name, "version": self.version}
 
 
+def normalize_discovery_date(value: Any) -> datetime.datetime:
+    if not isinstance(value, datetime.datetime):
+        raise TypeError("discovery_date must be a timezone-aware datetime.")
+
+    if value.tzinfo is None:
+        raise ValueError("discovery_date must be a timezone-aware datetime.")
+
+    # Normalize timezone to utc, and truncate microseconds to 0
+    return value.astimezone(datetime.timezone.utc).replace(microsecond=0)
+
+
 @attr.s(frozen=True, slots=True)
 class RawExtrinsicMetadata(HashableObject, BaseModel):
     object_type: Final = "raw_extrinsic_metadata"
@@ -893,7 +1213,7 @@ class RawExtrinsicMetadata(HashableObject, BaseModel):
     target = attr.ib(type=ExtendedSWHID, validator=type_validator())
 
     # source
-    discovery_date = attr.ib(type=datetime.datetime, validator=type_validator())
+    discovery_date = attr.ib(type=datetime.datetime, converter=normalize_discovery_date)
     authority = attr.ib(type=MetadataAuthority, validator=type_validator())
     fetcher = attr.ib(type=MetadataFetcher, validator=type_validator())
 
@@ -918,16 +1238,12 @@ class RawExtrinsicMetadata(HashableObject, BaseModel):
         type=Optional[CoreSWHID], default=None, validator=type_validator()
     )
 
-    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"")
+    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"", repr=hash_repr)
 
-    def compute_hash(self) -> bytes:
-        return hash_to_bytes(raw_extrinsic_metadata_identifier(self.to_dict()))
-
-    @discovery_date.validator
-    def check_discovery_date(self, attribute, value):
-        """Checks the discovery_date has a timezone."""
-        if value is not None and value.tzinfo is None:
-            raise ValueError("discovery_date must be a timezone-aware datetime.")
+    def _compute_hash_from_attributes(self) -> bytes:
+        return _compute_hash_from_manifest(
+            git_objects.raw_extrinsic_metadata_git_object(self)
+        )
 
     @origin.validator
     def check_origin(self, attribute, value):
@@ -1113,8 +1429,9 @@ class ExtID(HashableObject, BaseModel):
     extid_type = attr.ib(type=str, validator=type_validator())
     extid = attr.ib(type=bytes, validator=type_validator())
     target = attr.ib(type=CoreSWHID, validator=type_validator())
+    extid_version = attr.ib(type=int, validator=type_validator(), default=0)
 
-    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"")
+    id = attr.ib(type=Sha1Git, validator=type_validator(), default=b"", repr=hash_repr)
 
     @classmethod
     def from_dict(cls, d):
@@ -1122,7 +1439,8 @@ class ExtID(HashableObject, BaseModel):
             extid=d["extid"],
             extid_type=d["extid_type"],
             target=CoreSWHID.from_string(d["target"]),
+            extid_version=d.get("extid_version", 0),
         )
 
-    def compute_hash(self) -> bytes:
-        return hash_to_bytes(extid_identifier(self.to_dict()))
+    def _compute_hash_from_attributes(self) -> bytes:
+        return _compute_hash_from_manifest(git_objects.extid_git_object(self))
