@@ -3,33 +3,30 @@
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
 
+import collections
 import copy
 import datetime
+import hashlib
+from typing import Any, List, Optional, Tuple, Union
 
 import attr
 from attrs_strict import AttributeTypeError
+import dateutil
 from hypothesis import given
 from hypothesis.strategies import binary
 import pytest
 
-from swh.model.hashutil import MultiHash, hash_to_bytes, hash_to_hex
+from swh.model.collections import ImmutableDict
+from swh.model.from_disk import DentryPerms
+import swh.model.git_objects
+from swh.model.hashutil import MultiHash, hash_to_bytes
 import swh.model.hypothesis_strategies as strategies
-from swh.model.identifiers import (
-    CoreSWHID,
-    ExtendedSWHID,
-    ObjectType,
-    content_identifier,
-    directory_identifier,
-    origin_identifier,
-    raw_extrinsic_metadata_identifier,
-    release_identifier,
-    revision_identifier,
-    snapshot_identifier,
-)
+import swh.model.model
 from swh.model.model import (
     BaseModel,
     Content,
     Directory,
+    DirectoryEntry,
     MetadataAuthority,
     MetadataAuthorityType,
     MetadataFetcher,
@@ -43,15 +40,19 @@ from swh.model.model import (
     Revision,
     SkippedContent,
     Snapshot,
+    TargetType,
     Timestamp,
     TimestampWithTimezone,
+    type_validator,
 )
+import swh.model.swhids
+from swh.model.swhids import CoreSWHID, ExtendedSWHID, ObjectType
 from swh.model.tests.swh_model_data import TEST_OBJECTS
 from swh.model.tests.test_identifiers import (
-    content_example,
+    TS_DATETIMES,
+    TS_TIMEZONES,
     directory_example,
     metadata_example,
-    origin_example,
     release_example,
     revision_example,
     snapshot_example,
@@ -80,6 +81,203 @@ def test_todict_inverse_fromdict(objtype_and_obj):
     assert obj_as_dict == type(obj).from_dict(obj_as_dict).to_dict()
 
 
+@given(strategies.objects())
+def test_repr(objtype_and_obj):
+    """Checks every model object has a working repr(), and that it can be eval()uated
+    (so that printed objects can be copy-pasted to write test cases.)"""
+    (obj_type, obj) = objtype_and_obj
+
+    r = repr(obj)
+    env = {
+        "tzutc": lambda: datetime.timezone.utc,
+        "tzfile": dateutil.tz.tzfile,
+        "hash_to_bytes": hash_to_bytes,
+        **swh.model.swhids.__dict__,
+        **swh.model.model.__dict__,
+    }
+    assert eval(r, env) == obj
+
+
+@attr.s
+class Cls1:
+    pass
+
+
+@attr.s
+class Cls2(Cls1):
+    pass
+
+
+_custom_namedtuple = collections.namedtuple("_custom_namedtuple", "a b")
+
+
+class _custom_tuple(tuple):
+    pass
+
+
+# List of (type, valid_values, invalid_values)
+_TYPE_VALIDATOR_PARAMETERS: List[Tuple[Any, List[Any], List[Any]]] = [
+    # base types:
+    (
+        bool,
+        [True, False],
+        [-1, 0, 1, 42, 1000, None, "123", 0.0, (), ("foo",), ImmutableDict()],
+    ),
+    (
+        int,
+        [-1, 0, 1, 42, 1000, DentryPerms.directory, True, False],
+        [None, "123", 0.0, (), ImmutableDict()],
+    ),
+    (
+        float,
+        [-1.0, 0.0, 1.0, float("infinity"), float("NaN")],
+        [True, False, None, 1, "1.2", (), ImmutableDict()],
+    ),
+    (
+        bytes,
+        [b"", b"123"],
+        [None, bytearray(b"\x12\x34"), "123", 0, 123, (), (1, 2, 3), ImmutableDict()],
+    ),
+    (str, ["", "123"], [None, b"123", b"", 0, (), (1, 2, 3), ImmutableDict()]),
+    (None, [None], [b"", b"123", "", "foo", 0, 123, ImmutableDict(), float("NaN")]),
+    # unions:
+    (
+        Optional[int],
+        [None, -1, 0, 1, 42, 1000, DentryPerms.directory],
+        ["123", 0.0, (), ImmutableDict()],
+    ),
+    (
+        Optional[bytes],
+        [None, b"", b"123"],
+        ["123", "", 0, (), (1, 2, 3), ImmutableDict()],
+    ),
+    (
+        Union[str, bytes],
+        ["", "123", b"123", b""],
+        [None, 0, (), (1, 2, 3), ImmutableDict()],
+    ),
+    (
+        Union[str, bytes, None],
+        ["", "123", b"123", b"", None],
+        [0, (), (1, 2, 3), ImmutableDict()],
+    ),
+    # tuples
+    (
+        Tuple[str, str],
+        [("foo", "bar"), ("", ""), _custom_namedtuple("", ""), _custom_tuple(("", ""))],
+        [("foo",), ("foo", "bar", "baz"), ("foo", 42), (42, "foo")],
+    ),
+    (
+        Tuple[str, ...],
+        [
+            ("foo",),
+            ("foo", "bar"),
+            ("", ""),
+            ("foo", "bar", "baz"),
+            _custom_namedtuple("", ""),
+            _custom_tuple(("", "")),
+        ],
+        [("foo", 42), (42, "foo")],
+    ),
+    # composite generic:
+    (
+        Tuple[Union[str, int], Union[str, int]],
+        [("foo", "foo"), ("foo", 42), (42, "foo"), (42, 42)],
+        [("foo", b"bar"), (b"bar", "foo")],
+    ),
+    (
+        Union[Tuple[str, str], Tuple[int, int]],
+        [("foo", "foo"), (42, 42)],
+        [("foo", b"bar"), (b"bar", "foo"), ("foo", 42), (42, "foo")],
+    ),
+    (
+        Tuple[Tuple[bytes, bytes], ...],
+        [(), ((b"foo", b"bar"),), ((b"foo", b"bar"), (b"baz", b"qux"))],
+        [((b"foo", "bar"),), ((b"foo", b"bar"), ("baz", b"qux"))],
+    ),
+    # standard types:
+    (
+        datetime.datetime,
+        [
+            datetime.datetime(2021, 12, 15, 12, 59, 27),
+            datetime.datetime(2021, 12, 15, 12, 59, 27, tzinfo=datetime.timezone.utc),
+        ],
+        [None, 123],
+    ),
+    # ImmutableDict
+    (
+        ImmutableDict[str, int],
+        [
+            ImmutableDict(),
+            ImmutableDict({"foo": 42}),
+            ImmutableDict({"foo": 42, "bar": 123}),
+        ],
+        [ImmutableDict({"foo": "bar"}), ImmutableDict({42: 123})],
+    ),
+    # Any:
+    (object, [-1, 0, 1, 42, 1000, None, "123", 0.0, (), ImmutableDict()], [],),
+    (Any, [-1, 0, 1, 42, 1000, None, "123", 0.0, (), ImmutableDict()], [],),
+    (
+        ImmutableDict[Any, int],
+        [
+            ImmutableDict(),
+            ImmutableDict({"foo": 42}),
+            ImmutableDict({"foo": 42, "bar": 123}),
+            ImmutableDict({42: 123}),
+        ],
+        [ImmutableDict({"foo": "bar"})],
+    ),
+    (
+        ImmutableDict[str, Any],
+        [
+            ImmutableDict(),
+            ImmutableDict({"foo": 42}),
+            ImmutableDict({"foo": "bar"}),
+            ImmutableDict({"foo": 42, "bar": 123}),
+        ],
+        [ImmutableDict({42: 123})],
+    ),
+    # attr objects:
+    (
+        Timestamp,
+        [Timestamp(seconds=123, microseconds=0),],
+        [None, "2021-09-28T11:27:59", 123],
+    ),
+    (Cls1, [Cls1(), Cls2()], [None, b"abcd"],),
+    # enums:
+    (
+        TargetType,
+        [TargetType.CONTENT, TargetType.ALIAS],
+        ["content", "alias", 123, None],
+    ),
+]
+
+
+@pytest.mark.parametrize(
+    "type_,value",
+    [
+        pytest.param(type_, value, id=f"type={type_}, value={value}")
+        for (type_, values, _) in _TYPE_VALIDATOR_PARAMETERS
+        for value in values
+    ],
+)
+def test_type_validator_valid(type_, value):
+    type_validator()(None, attr.ib(type=type_), value)
+
+
+@pytest.mark.parametrize(
+    "type_,value",
+    [
+        pytest.param(type_, value, id=f"type={type_}, value={value}")
+        for (type_, _, values) in _TYPE_VALIDATOR_PARAMETERS
+        for value in values
+    ],
+)
+def test_type_validator_invalid(type_, value):
+    with pytest.raises(AttributeTypeError):
+        type_validator()(None, attr.ib(type=type_), value)
+
+
 @pytest.mark.parametrize("object_type, objects", TEST_OBJECTS.items())
 def test_swh_model_todict_fromdict(object_type, objects):
     """checks model objects in swh_model_data are in correct shape"""
@@ -255,21 +453,27 @@ def test_timestampwithtimezone():
     tstz = TimestampWithTimezone(timestamp=ts, offset=0, negative_utc=False)
     attr.validate(tstz)
     assert tstz.negative_utc is False
+    assert tstz.offset_bytes == b"+0000"
 
-    attr.validate(TimestampWithTimezone(timestamp=ts, offset=10, negative_utc=False))
+    tstz = TimestampWithTimezone(timestamp=ts, offset=10, negative_utc=False)
+    attr.validate(tstz)
+    assert tstz.offset_bytes == b"+0010"
 
-    attr.validate(TimestampWithTimezone(timestamp=ts, offset=-10, negative_utc=False))
+    tstz = TimestampWithTimezone(timestamp=ts, offset=-10, negative_utc=False)
+    attr.validate(tstz)
+    assert tstz.offset_bytes == b"-0010"
 
     tstz = TimestampWithTimezone(timestamp=ts, offset=0, negative_utc=True)
     attr.validate(tstz)
     assert tstz.negative_utc is True
+    assert tstz.offset_bytes == b"-0000"
 
     with pytest.raises(AttributeTypeError):
         TimestampWithTimezone(
             timestamp=datetime.datetime.now(), offset=0, negative_utc=False
         )
 
-    with pytest.raises(AttributeTypeError):
+    with pytest.raises((AttributeTypeError, TypeError)):
         TimestampWithTimezone(timestamp=ts, offset="0", negative_utc=False)
 
     with pytest.raises(AttributeTypeError):
@@ -329,6 +533,17 @@ def test_timestampwithtimezone_from_iso8601_negative_utc():
     )
 
 
+@pytest.mark.parametrize("date", TS_DATETIMES)
+@pytest.mark.parametrize("tz", TS_TIMEZONES)
+@pytest.mark.parametrize("microsecond", [0, 1, 10, 100, 1000, 999999])
+def test_timestampwithtimezone_to_datetime(date, tz, microsecond):
+    date = date.replace(tzinfo=tz, microsecond=microsecond)
+    tstz = TimestampWithTimezone.from_datetime(date)
+
+    assert tstz.to_datetime() == date
+    assert tstz.to_datetime().utcoffset() == date.utcoffset()
+
+
 def test_person_from_fullname():
     """The author should have name, email and fullname filled.
 
@@ -535,9 +750,141 @@ def test_skipped_content_naive_datetime():
         )
 
 
+# Directory
+
+
+@given(strategies.directories().filter(lambda d: d.raw_manifest is None))
+def test_directory_check(directory):
+    directory.check()
+
+    directory2 = attr.evolve(directory, id=b"\x00" * 20)
+    with pytest.raises(ValueError, match="does not match recomputed hash"):
+        directory2.check()
+
+    directory2 = attr.evolve(
+        directory, raw_manifest=swh.model.git_objects.directory_git_object(directory)
+    )
+    with pytest.raises(
+        ValueError, match="non-none raw_manifest attribute, but does not need it."
+    ):
+        directory2.check()
+
+
+@given(strategies.directories().filter(lambda d: d.raw_manifest is None))
+def test_directory_raw_manifest(directory):
+    assert "raw_manifest" not in directory.to_dict()
+
+    raw_manifest = b"foo"
+    id_ = hashlib.new("sha1", raw_manifest).digest()
+
+    directory2 = attr.evolve(directory, raw_manifest=raw_manifest)
+    assert directory2.to_dict()["raw_manifest"] == raw_manifest
+    with pytest.raises(ValueError, match="does not match recomputed hash"):
+        directory2.check()
+
+    directory2 = attr.evolve(directory, raw_manifest=raw_manifest, id=id_)
+    assert directory2.id is not None
+    assert directory2.id == id_ != directory.id
+    assert directory2.to_dict()["raw_manifest"] == raw_manifest
+    directory2.check()
+
+
+def test_directory_entry_name_validation():
+    with pytest.raises(ValueError, match="valid directory entry name."):
+        DirectoryEntry(name=b"foo/", type="dir", target=b"\x00" * 20, perms=0),
+
+
+def test_directory_duplicate_entry_name():
+    entries = (
+        DirectoryEntry(name=b"foo", type="file", target=b"\x00" * 20, perms=0),
+        DirectoryEntry(name=b"foo", type="dir", target=b"\x01" * 20, perms=1),
+    )
+    with pytest.raises(ValueError, match="duplicated entry name"):
+        Directory(entries=entries)
+
+    entries = (
+        DirectoryEntry(name=b"foo", type="file", target=b"\x00" * 20, perms=0),
+        DirectoryEntry(name=b"foo", type="file", target=b"\x00" * 20, perms=0),
+    )
+    with pytest.raises(ValueError, match="duplicated entry name"):
+        Directory(entries=entries)
+
+
+# Release
+
+
+@given(strategies.releases().filter(lambda rel: rel.raw_manifest is None))
+def test_release_check(release):
+    release.check()
+
+    release2 = attr.evolve(release, id=b"\x00" * 20)
+    with pytest.raises(ValueError, match="does not match recomputed hash"):
+        release2.check()
+
+    release2 = attr.evolve(
+        release, raw_manifest=swh.model.git_objects.release_git_object(release)
+    )
+    with pytest.raises(
+        ValueError, match="non-none raw_manifest attribute, but does not need it."
+    ):
+        release2.check()
+
+
+@given(strategies.releases().filter(lambda rev: rev.raw_manifest is None))
+def test_release_raw_manifest(release):
+    raw_manifest = b"foo"
+    id_ = hashlib.new("sha1", raw_manifest).digest()
+
+    release2 = attr.evolve(release, raw_manifest=raw_manifest)
+    assert release2.to_dict()["raw_manifest"] == raw_manifest
+    with pytest.raises(ValueError, match="does not match recomputed hash"):
+        release2.check()
+
+    release2 = attr.evolve(release, raw_manifest=raw_manifest, id=id_)
+    assert release2.id is not None
+    assert release2.id == id_ != release.id
+    assert release2.to_dict()["raw_manifest"] == raw_manifest
+    release2.check()
+
+
 # Revision
 
 
+@given(strategies.revisions().filter(lambda rev: rev.raw_manifest is None))
+def test_revision_check(revision):
+    revision.check()
+
+    revision2 = attr.evolve(revision, id=b"\x00" * 20)
+    with pytest.raises(ValueError, match="does not match recomputed hash"):
+        revision2.check()
+
+    revision2 = attr.evolve(
+        revision, raw_manifest=swh.model.git_objects.revision_git_object(revision)
+    )
+    with pytest.raises(
+        ValueError, match="non-none raw_manifest attribute, but does not need it."
+    ):
+        revision2.check()
+
+
+@given(strategies.revisions().filter(lambda rev: rev.raw_manifest is None))
+def test_revision_raw_manifest(revision):
+
+    raw_manifest = b"foo"
+    id_ = hashlib.new("sha1", raw_manifest).digest()
+
+    revision2 = attr.evolve(revision, raw_manifest=raw_manifest)
+    assert revision2.to_dict()["raw_manifest"] == raw_manifest
+    with pytest.raises(ValueError, match="does not match recomputed hash"):
+        revision2.check()
+
+    revision2 = attr.evolve(revision, raw_manifest=raw_manifest, id=id_)
+    assert revision2.id is not None
+    assert revision2.id == id_ != revision.id
+    assert revision2.to_dict()["raw_manifest"] == raw_manifest
+    revision2.check()
+
+
 def test_revision_extra_headers_no_headers():
     rev_dict = revision_example.copy()
     rev_dict.pop("id")
@@ -723,94 +1070,6 @@ def test_revision_extra_headers_as_lists_from_dict():
     assert rev_model.extra_headers == extra_headers
 
 
-# ID computation
-
-
-def test_content_model_id_computation():
-    cnt_dict = content_example.copy()
-
-    cnt_id_str = hash_to_hex(content_identifier(cnt_dict)["sha1_git"])
-    cnt_model = Content.from_data(cnt_dict["data"])
-    assert str(cnt_model.swhid()) == "swh:1:cnt:" + cnt_id_str
-
-
-def test_directory_model_id_computation():
-    dir_dict = directory_example.copy()
-    del dir_dict["id"]
-
-    dir_id_str = directory_identifier(dir_dict)
-    dir_id = hash_to_bytes(dir_id_str)
-    dir_model = Directory.from_dict(dir_dict)
-    assert dir_model.id == dir_id
-    assert str(dir_model.swhid()) == "swh:1:dir:" + dir_id_str
-
-
-def test_revision_model_id_computation():
-    rev_dict = revision_example.copy()
-    del rev_dict["id"]
-
-    rev_id_str = revision_identifier(rev_dict)
-    rev_id = hash_to_bytes(rev_id_str)
-    rev_model = Revision.from_dict(rev_dict)
-    assert rev_model.id == rev_id
-    assert str(rev_model.swhid()) == "swh:1:rev:" + rev_id_str
-
-
-def test_revision_model_id_computation_with_no_date():
-    """We can have revision with date to None
-
-    """
-    rev_dict = revision_example.copy()
-    rev_dict["date"] = None
-    rev_dict["committer_date"] = None
-    del rev_dict["id"]
-
-    rev_id = hash_to_bytes(revision_identifier(rev_dict))
-    rev_model = Revision.from_dict(rev_dict)
-    assert rev_model.date is None
-    assert rev_model.committer_date is None
-    assert rev_model.id == rev_id
-
-
-def test_release_model_id_computation():
-    rel_dict = release_example.copy()
-    del rel_dict["id"]
-
-    rel_id_str = release_identifier(rel_dict)
-    rel_id = hash_to_bytes(rel_id_str)
-    rel_model = Release.from_dict(rel_dict)
-    assert isinstance(rel_model.date, TimestampWithTimezone)
-    assert rel_model.id == hash_to_bytes(rel_id)
-    assert str(rel_model.swhid()) == "swh:1:rel:" + rel_id_str
-
-
-def test_snapshot_model_id_computation():
-    snp_dict = snapshot_example.copy()
-    del snp_dict["id"]
-
-    snp_id_str = snapshot_identifier(snp_dict)
-    snp_id = hash_to_bytes(snp_id_str)
-    snp_model = Snapshot.from_dict(snp_dict)
-    assert snp_model.id == snp_id
-    assert str(snp_model.swhid()) == "swh:1:snp:" + snp_id_str
-
-
-def test_origin_model_id_computation():
-    ori_dict = origin_example.copy()
-
-    ori_id_str = origin_identifier(ori_dict)
-    ori_model = Origin.from_dict(ori_dict)
-    assert str(ori_model.swhid()) == "swh:1:ori:" + ori_id_str
-
-
-def test_raw_extrinsic_metadata_model_id_computation():
-    emd_dict = metadata_example.copy()
-
-    emd_id_str = raw_extrinsic_metadata_identifier(emd_dict)
-    emd_model = RawExtrinsicMetadata.from_dict(emd_dict)
-    assert str(emd_model.swhid()) == "swh:1:emd:" + emd_id_str
-
-
 @given(strategies.objects(split_content=True))
 def test_object_type(objtype_and_obj):
     obj_type, obj = objtype_and_obj
@@ -1140,3 +1399,39 @@ def test_metadata_validate_context_directory():
             ),
             **_common_metadata_fields,
         )
+
+
+def test_metadata_normalize_discovery_date():
+    fields_copy = {**_common_metadata_fields}
+    truncated_date = fields_copy.pop("discovery_date")
+    assert truncated_date.microsecond == 0
+
+    # Check for TypeError on disabled object type: we removed attrs_strict's
+    # type_validator
+    with pytest.raises(TypeError):
+        RawExtrinsicMetadata(
+            target=_content_swhid, discovery_date="not a datetime", **fields_copy
+        )
+
+    # Check for truncation to integral second
+    date_with_us = truncated_date.replace(microsecond=42)
+    md = RawExtrinsicMetadata(
+        target=_content_swhid, discovery_date=date_with_us, **fields_copy,
+    )
+
+    assert md.discovery_date == truncated_date
+    assert md.discovery_date.tzinfo == datetime.timezone.utc
+
+    # Check that the timezone gets normalized. Timezones can be offset by a
+    # non-integral number of seconds, so we need to handle that.
+    timezone = datetime.timezone(offset=datetime.timedelta(hours=2))
+    date_with_tz = truncated_date.astimezone(timezone)
+
+    assert date_with_tz.tzinfo != datetime.timezone.utc
+
+    md = RawExtrinsicMetadata(
+        target=_content_swhid, discovery_date=date_with_tz, **fields_copy,
+    )
+
+    assert md.discovery_date == truncated_date
+    assert md.discovery_date.tzinfo == datetime.timezone.utc
